# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import sys

from twisted.internet import defer

from buildbot.config import BuilderConfig
from buildbot.plugins import schedulers
from buildbot.plugins import steps
from buildbot.process.factory import BuildFactory
from buildbot.test.util.integration import RunMasterBase
from buildbot.util import asyncSleep


# This integration test creates a master and worker environment,
# with one builders and a shellcommand step
# meant to be a template for integration steps
class ShellMaster(RunMasterBase):

    def config_for_master_command(self, **kwargs):
        c = {}

        c['schedulers'] = [
            schedulers.AnyBranchScheduler(name="sched", builderNames=["testy"])
        ]

        f = BuildFactory()
        f.addStep(steps.MasterShellCommand(**kwargs))
        c['builders'] = [
            BuilderConfig(name="testy", workernames=["local1"], factory=f)
        ]
        return c

    def get_change(self):
        return {
            "branch": "master",
            "files": ["foo.c"],
            "author": "me@foo.com",
            "committer": "me@foo.com",
            "comments": "good stuff",
            "revision": "HEAD",
            "project": "none"
        }

    @defer.inlineCallbacks
    def test_shell(self):
        yield self.setupConfig(self.config_for_master_command(command='echo hello'))

        build = yield self.doForceBuild(wantSteps=True, useChange=self.get_change(), wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        self.assertEqual(build['steps'][1]['state_string'], 'Ran')

    @defer.inlineCallbacks
    def test_logs(self):
        yield self.setupConfig(self.config_for_master_command(command=[
            sys.executable, '-c', 'print("hello")'
        ]))

        build = yield self.doForceBuild(wantSteps=True, useChange=self.get_change(), wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        res = yield self.checkBuildStepLogExist(build, "hello")
        self.assertTrue(res)
        self.assertEqual(build['steps'][1]['state_string'], 'Ran')

    @defer.inlineCallbacks
    def test_fails(self):
        yield self.setupConfig(self.config_for_master_command(command=[
            sys.executable, '-c', 'exit(1)'
        ]))

        build = yield self.doForceBuild(wantSteps=True, useChange=self.get_change(), wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        self.assertEqual(build['steps'][1]['state_string'], 'failed (1) (failure)')

    @defer.inlineCallbacks
    def test_interrupt(self):
        yield self.setupConfig(self.config_for_master_command(name='sleep', command=[
            sys.executable, '-c', "while True: pass"
        ]))

        d = self.doForceBuild(wantSteps=True, useChange=self.get_change(), wantLogs=True)

        @defer.inlineCallbacks
        def on_new_step(_, data):
            if data['name'] == 'sleep':
                # wait until the step really starts
                yield asyncSleep(1)
                brs = yield self.master.data.get(('buildrequests',))
                brid = brs[-1]['buildrequestid']
                self.master.data.control('cancel', {'reason': 'cancelled by test'},
                                         ('buildrequests', brid))

        yield self.master.mq.startConsuming(on_new_step, ('steps', None, 'new'))

        build = yield d
        self.assertEqual(build['buildid'], 1)
        if sys.platform == 'win32':
            self.assertEqual(build['steps'][1]['state_string'], 'failed (1) (exception)')
        else:
            self.assertEqual(build['steps'][1]['state_string'], 'killed (9) (exception)')



# coding=utf-8
# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""
Helpers for handling compatibility differences
between Python 2 and Python 3.
"""

from __future__ import absolute_import
from __future__ import print_function
from future.utils import text_type

if str != bytes:
    # On Python 3 and higher, str and bytes
    # are not equivalent.  We must use StringIO for
    # doing io on native strings.
    from io import StringIO as NativeStringIO
else:
    # On Python 2 and older, str and bytes
    # are equivalent.  We must use BytesIO for
    # doing io on native strings.
    from io import BytesIO as NativeStringIO


def bytes2NativeString(x, encoding='utf-8'):
    """
    Convert C{bytes} to a native C{str}.

    On Python 3 and higher, str and bytes
    are not equivalent.  In this case, decode
    the bytes, and return a native string.

    On Python 2 and lower, str and bytes
    are equivalent.  In this case, just
    just return the native string.

    @param x: a string of type C{bytes}
    @param encoding: an optional codec, default: 'utf-8'
    @return: a string of type C{str}
    """
    if isinstance(x, bytes) and str != bytes:
        return x.decode(encoding)
    return x


def unicode2bytes(x, encoding='utf-8', errors='strict'):
    """
    Convert a unicode string to C{bytes}.

    @param x: a unicode string, of type C{unicode} on Python 2,
              or C{str} on Python 3.
    @param encoding: an optional codec, default: 'utf-8'
    @param errors: error handling scheme, default 'strict'
    @return: a string of type C{bytes}
    """
    if isinstance(x, text_type):
        x = x.encode(encoding, errors)
    return x


def bytes2unicode(x, encoding='utf-8', errors='strict'):
    """
    Convert a C{bytes} to a unicode string.

    @param x: a unicode string, of type C{unicode} on Python 2,
              or C{str} on Python 3.
    @param encoding: an optional codec, default: 'utf-8'
    @param errors: error handling scheme, default 'strict'
    @return: a unicode string of type C{unicode} on Python 2, or
             C{str} on Python 3.
    """
    if isinstance(x, (text_type, type(None))):
        return x
    return text_type(x, encoding, errors)


__all__ = [
    "NativeStringIO",
    "bytes2NativeString",
    "bytes2unicode",
    "unicode2bytes"
]



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.test.util import interfaces


class TestAssertArgSpecMatches(interfaces.InterfaceTests, unittest.TestCase):

    def test_simple_decorator(self):
        def myfunc(x, y=2, *args):
            pass

        @self.assertArgSpecMatches(myfunc)
        def myfunc2(x, y=2, *args):
            pass

        try:
            @self.assertArgSpecMatches(myfunc)
            def myfunc3(x, y=3, *args):
                pass
        except Exception as e:
            error = e
        else:
            error = None

        self.assertIdentical(type(error), unittest.FailTest)
        self.assertEqual(
            error.args, ('Expected: (x, y=3, *args); got: (x, y=2, *args)',))

    def test_double_decorator(self):
        def myfunc(x, y):
            pass

        def myfunc2(x, y):
            pass

        def myfunc3(x, yy):
            pass

        @self.assertArgSpecMatches(myfunc, myfunc2)
        def myfunc4(x, y):
            pass

        try:
            @self.assertArgSpecMatches(myfunc, myfunc3)
            def myfunc5(x, y):
                pass
        except Exception as e:
            error = e
        else:
            error = None

        self.assertIdentical(type(error), unittest.FailTest)
        self.assertEqual(error.args, ('Expected: (x, y); got: (x, yy)',))

        try:
            @self.assertArgSpecMatches(myfunc, myfunc3)
            def myfunc6(xx, yy):
                pass
        except Exception as e:
            error = e
        else:
            error = None

        self.assertIdentical(type(error), unittest.FailTest)
        self.assertEqual(error.args, ('Expected: (x, y); got: (x, yy)',))

    def test_function_style(self):
        def myfunc(x, y=2, *args):
            pass

        def myfunc2(x, y=2, *args):
            pass

        def myfunc3(x, y=3, *args):
            pass

        self.assertArgSpecMatches(myfunc, myfunc2)

        try:
            self.assertArgSpecMatches(myfunc, myfunc3)
        except Exception as e:
            error = e
        else:
            error = None

        self.assertIdentical(type(error), unittest.FailTest)
        self.assertEqual(
            error.args, ('Expected: (x, y=2, *args); got: (x, y=3, *args)',))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
Test clean shutdown functionality of the master
"""

import mock

from twisted.cred import credentials
from twisted.internet import defer
from twisted.spread import pb
from twisted.trial import unittest

from buildbot.worker.protocols.manager.pb import PBManager


class FakeMaster:
    initLock = defer.DeferredLock()

    def addService(self, svc):
        pass

    @property
    def master(self):
        return self


class TestPBManager(unittest.TestCase):

    @defer.inlineCallbacks
    def setUp(self):
        self.pbm = PBManager()
        yield self.pbm.setServiceParent(FakeMaster())
        self.pbm.startService()
        self.connections = []

    def tearDown(self):
        return self.pbm.stopService()

    def perspectiveFactory(self, mind, username):
        persp = mock.Mock()
        persp.is_my_persp = True
        persp.attached = lambda mind: defer.succeed(None)
        self.connections.append(username)
        return defer.succeed(persp)

    @defer.inlineCallbacks
    def test_register_unregister(self):
        portstr = "tcp:0:interface=127.0.0.1"
        reg = yield self.pbm.register(portstr, "boris", "pass", self.perspectiveFactory)

        # make sure things look right
        self.assertIn(portstr, self.pbm.dispatchers)
        disp = self.pbm.dispatchers[portstr]
        self.assertIn('boris', disp.users)

        # we can't actually connect to it, as that requires finding the
        # dynamically allocated port number which is buried out of reach;
        # however, we can try the requestAvatar and requestAvatarId methods.

        username = yield disp.requestAvatarId(credentials.UsernamePassword(b'boris', b'pass'))

        self.assertEqual(username, b'boris')
        avatar = yield disp.requestAvatar(b'boris', mock.Mock(), pb.IPerspective)

        _, persp, __ = avatar
        self.assertTrue(persp.is_my_persp)
        self.assertIn('boris', self.connections)

        yield reg.unregister()

    @defer.inlineCallbacks
    def test_register_no_user(self):
        portstr = "tcp:0:interface=127.0.0.1"
        reg = yield self.pbm.register(portstr, "boris", "pass", self.perspectiveFactory)

        # make sure things look right
        self.assertIn(portstr, self.pbm.dispatchers)
        disp = self.pbm.dispatchers[portstr]
        self.assertIn('boris', disp.users)

        # we can't actually connect to it, as that requires finding the
        # dynamically allocated port number which is buried out of reach;
        # however, we can try the requestAvatar and requestAvatarId methods.

        username = yield disp.requestAvatarId(credentials.UsernamePassword(b'boris', b'pass'))

        self.assertEqual(username, b'boris')
        with self.assertRaises(ValueError):
            yield disp.requestAvatar(b'notboris', mock.Mock(), pb.IPerspective)

        self.assertNotIn('boris', self.connections)

        yield reg.unregister()

    @defer.inlineCallbacks
    def test_requestAvatarId_noinitLock(self):
        portstr = "tcp:0:interface=127.0.0.1"
        reg = yield self.pbm.register(portstr, "boris", "pass", self.perspectiveFactory)

        disp = self.pbm.dispatchers[portstr]

        d = disp.requestAvatarId(credentials.UsernamePassword(b'boris', b'pass'))
        self.assertTrue(d.called,
            "requestAvatarId should have been called since the lock is free")

        yield reg.unregister()

    @defer.inlineCallbacks
    def test_requestAvatarId_initLock(self):
        portstr = "tcp:0:interface=127.0.0.1"
        reg = yield self.pbm.register(portstr, "boris", "pass", self.perspectiveFactory)

        disp = self.pbm.dispatchers[portstr]

        try:
            # simulate a reconfig/restart in progress
            yield self.pbm.master.initLock.acquire()
            # try to authenticate while the lock is locked
            d = disp.requestAvatarId(credentials.UsernamePassword(b'boris', b'pass'))
            self.assertFalse(d.called,
                "requestAvatarId should block until the lock is released")
        finally:
            # release the lock, it should allow for auth to proceed
            yield self.pbm.master.initLock.release()

        self.assertTrue(d.called,
            "requestAvatarId should have been called after the lock was released")
        yield reg.unregister()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.python import log
from zope.interface import implementer

from buildbot import config
from buildbot.interfaces import IChangeSource
from buildbot.util import service
from buildbot.util.poll import method as poll_method
from buildbot.warnings import warn_deprecated


@implementer(IChangeSource)
class ChangeSource(service.ClusteredBuildbotService):

    def describe(self):
        pass

    # activity handling

    def activate(self):
        return defer.succeed(None)

    def deactivate(self):
        return defer.succeed(None)

    # service handling

    def _getServiceId(self):
        return self.master.data.updates.findChangeSourceId(self.name)

    def _claimService(self):
        return self.master.data.updates.trySetChangeSourceMaster(self.serviceid,
                                                                 self.master.masterid)

    def _unclaimService(self):
        return self.master.data.updates.trySetChangeSourceMaster(self.serviceid,
                                                                 None)


class ReconfigurablePollingChangeSource(ChangeSource):
    pollInterval = None
    pollAtLaunch = None
    pollRandomDelayMin = None
    pollRandomDelayMax = None

    def checkConfig(self, name=None, pollInterval=60 * 10, pollAtLaunch=False,
                    pollRandomDelayMin=0, pollRandomDelayMax=0):
        super().checkConfig(name=name)
        if pollInterval < 0:
            config.error(f"interval must be >= 0: {pollInterval}")
        if pollRandomDelayMin < 0:
            config.error(f"min random delay must be >= 0: {pollRandomDelayMin}")
        if pollRandomDelayMax < 0:
            config.error(f"max random delay must be >= 0: {pollRandomDelayMax}")
        if pollRandomDelayMin > pollRandomDelayMax:
            config.error(f"min random delay must be <= {pollRandomDelayMax}: {pollRandomDelayMin}")
        if pollRandomDelayMax >= pollInterval:
            config.error(f"max random delay must be < {pollInterval}: {pollRandomDelayMax}")

    @defer.inlineCallbacks
    def reconfigService(self, name=None, pollInterval=60 * 10, pollAtLaunch=False,
                        pollRandomDelayMin=0, pollRandomDelayMax=0):
        self.pollInterval, prevPollInterval = pollInterval, self.pollInterval
        self.pollAtLaunch = pollAtLaunch
        self.pollRandomDelayMin = pollRandomDelayMin
        self.pollRandomDelayMax = pollRandomDelayMax
        yield super().reconfigService(name=name)

        # pollInterval change is the only value which makes sense to reconfigure check.
        if prevPollInterval != pollInterval and self.doPoll.running:
            yield self.doPoll.stop()
            # As a implementation detail, poller will 'pollAtReconfigure' if poll interval changes
            # and pollAtLaunch=True
            yield self.doPoll.start(interval=self.pollInterval, now=self.pollAtLaunch,
                                    random_delay_min=self.pollRandomDelayMin,
                                    random_delay_max=self.pollRandomDelayMax)

    def poll(self):
        pass

    @poll_method
    def doPoll(self):
        d = defer.maybeDeferred(self.poll)
        d.addErrback(log.err, f'{self}: while polling for changes')
        return d

    def force(self):
        self.doPoll()

    def activate(self):
        self.doPoll.start(interval=self.pollInterval, now=self.pollAtLaunch,
                          random_delay_min=self.pollRandomDelayMin,
                          random_delay_max=self.pollRandomDelayMax)

    def deactivate(self):
        return self.doPoll.stop()


class PollingChangeSource(ReconfigurablePollingChangeSource):
    # Legacy code will be very painful to port to BuildbotService life cycle
    # because the unit tests keep doing shortcuts for the Service life cycle (i.e by no calling
    # startService) instead of porting everything at once, we make a class to support legacy

    def checkConfig(self, name=None, pollInterval=60 * 10, pollAtLaunch=False,
                    pollRandomDelayMin=0, pollRandomDelayMax=0, **kwargs):
        super().checkConfig(name=name, pollInterval=60 * 10, pollAtLaunch=False,
                            pollRandomDelayMin=0, pollRandomDelayMax=0)

        warn_deprecated('3.3.0', 'PollingChangeSource has been deprecated: ' +
                        'please use ReconfigurablePollingChangeSource')

        self.pollInterval = pollInterval
        self.pollAtLaunch = pollAtLaunch
        self.pollRandomDelayMin = pollRandomDelayMin
        self.pollRandomDelayMax = pollRandomDelayMax

    def reconfigService(self, *args, **kwargs):
        # BuildbotServiceManager will detect such exception and swap old service with new service,
        # instead of just reconfiguring
        raise NotImplementedError()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import re

from parameterized import parameterized

from twisted.trial import unittest

from buildbot.changes.filter import ChangeFilter
from buildbot.test.fake.change import Change


class TestChangeFilter(unittest.TestCase):
    def test_filter_change_filter_fn(self):
        f = ChangeFilter(filter_fn=lambda ch: ch.x > 3)
        self.assertFalse(f.filter_change(Change(x=2)))
        self.assertTrue(f.filter_change(Change(x=4)))

        self.assertEqual(repr(f), "<ChangeFilter on <lambda>()>")

    test_cases = [
        (
            "match",
            Change(project="p", codebase="c", repository="r", category="ct", branch="b"),
            True,
        ),
        (
            "not_project",
            Change(project="p0", codebase="c", repository="r", category="ct", branch="b"),
            False,
        ),
        (
            "not_codebase",
            Change(project="p", codebase="c0", repository="r", category="ct", branch="b"),
            False,
        ),
        (
            "not_repository",
            Change(project="p", codebase="c", repository="r0", category="ct", branch="b"),
            False,
        ),
        (
            "not_category",
            Change(project="p", codebase="c", repository="r", category="ct0", branch="b"),
            False,
        ),
        (
            "not_branch",
            Change(project="p", codebase="c", repository="r", category="ct", branch="b0"),
            False,
        ),
    ]

    @parameterized.expand(test_cases)
    def test_eq(self, name, change, expected):
        f = ChangeFilter(project="p", codebase="c", repository="r", category="ct", branch="b")
        self.assertEqual(f.filter_change(change), expected)

    @parameterized.expand(test_cases)
    def test_eq_list(self, name, change, expected):
        f = ChangeFilter(
            project=["p", "p9"],
            codebase=["c", "c9"],
            repository=["r", "r9"],
            category=["ct", "ct9"],
            branch=["b", "b9"],
        )
        self.assertEqual(f.filter_change(change), expected)

    @parameterized.expand(test_cases)
    def test_not_eq(self, name, change, expected):
        f = ChangeFilter(
            project_not_eq="p0",
            codebase_not_eq="c0",
            repository_not_eq="r0",
            category_not_eq="ct0",
            branch_not_eq="b0",
        )
        self.assertEqual(f.filter_change(change), expected)

    @parameterized.expand(test_cases)
    def test_not_eq_list(self, name, change, expected):
        f = ChangeFilter(
            project_not_eq=["p0", "p1"],
            codebase_not_eq=["c0", "c1"],
            repository_not_eq=["r0", "r1"],
            category_not_eq=["ct0", "ct1"],
            branch_not_eq=["b0", "b1"],
        )
        self.assertEqual(f.filter_change(change), expected)

    @parameterized.expand(test_cases)
    def test_re(self, name, change, expected):
        f = ChangeFilter(
            project_re="^p$",
            codebase_re="^c$",
            repository_re="^r$",
            category_re="^ct$",
            branch_re="^b$",
        )
        self.assertEqual(f.filter_change(change), expected)

    @parameterized.expand(test_cases)
    def test_re_list(self, name, change, expected):
        f = ChangeFilter(
            project_re=["^p$", "^p1$"],
            codebase_re=["^c$", "^c1$"],
            repository_re=["^r$", "^r1$"],
            category_re=["^ct$", "^ct1$"],
            branch_re=["^b$", "^b1$"],
        )
        self.assertEqual(f.filter_change(change), expected)

    @parameterized.expand(test_cases)
    def test_not_re(self, name, change, expected):
        f = ChangeFilter(
            project_not_re="^p0$",
            codebase_not_re="^c0$",
            repository_not_re="^r0$",
            category_not_re="^ct0$",
            branch_not_re="^b0$",
        )
        self.assertEqual(f.filter_change(change), expected)

    @parameterized.expand(test_cases)
    def test_not_re_list(self, name, change, expected):
        f = ChangeFilter(
            project_not_re=["^p0$", "^p1$"],
            codebase_not_re=["^c0$", "^c1$"],
            repository_not_re=["^r0$", "^r1$"],
            category_not_re=["^ct0$", "^ct1$"],
            branch_not_re=["^b0$", "^b1$"],
        )
        self.assertEqual(f.filter_change(change), expected)

    @parameterized.expand(test_cases)
    def test_re_pattern(self, name, change, expected):
        f = ChangeFilter(
            project_re=re.compile("^p$"),
            codebase_re=re.compile("^c$"),
            repository_re=re.compile("^r$"),
            category_re=re.compile("^ct$"),
            branch_re=re.compile("^b$"),
        )
        self.assertEqual(f.filter_change(change), expected)

    @parameterized.expand(test_cases)
    def test_fn(self, name, change, expected):
        f = ChangeFilter(
            project_fn=lambda p: p == "p",
            codebase_fn=lambda p: p == "c",
            repository_fn=lambda p: p == "r",
            category_fn=lambda p: p == "ct",
            branch_fn=lambda p: p == "b",
        )
        self.assertEqual(f.filter_change(change), expected)

        self.assertEqual(
            repr(f),
            "<ChangeFilter on <lambda>(project) and <lambda>(codebase) and "
            "<lambda>(repository) and <lambda>(category) and <lambda>(branch)>",
        )

    def test_filter_change_filt_branch_list_None(self):
        f = ChangeFilter(branch=["mybr", None])
        self.assertTrue(f.filter_change(Change(branch="mybr")))
        self.assertTrue(f.filter_change(Change(branch=None)))
        self.assertFalse(f.filter_change(Change(branch="misc")))

    def test_filter_change_branch_re(self):  # regression - see #927
        f = ChangeFilter(branch_re="^t.*")
        self.assertTrue(f.filter_change(Change(branch="trunk")))
        self.assertFalse(f.filter_change(Change(branch="development")))
        self.assertFalse(f.filter_change(Change(branch=None)))

    def test_filter_change_combination(self):
        f = ChangeFilter(project="p", repository="r", branch="b", category="c", codebase="cb")
        self.assertFalse(
            f.filter_change(Change(project="x", repository="x", branch="x", category="x"))
        )
        self.assertFalse(
            f.filter_change(Change(project="p", repository="r", branch="b", category="x"))
        )
        self.assertFalse(
            f.filter_change(Change(project="p", repository="r", branch="b", category="c"))
        )
        self.assertTrue(
            f.filter_change(
                Change(project="p", repository="r", branch="b", category="c", codebase="cb")
            )
        )

    def test_filter_change_combination_filter_fn(self):
        f = ChangeFilter(
            project="p",
            repository="r",
            branch="b",
            category="c",
            filter_fn=lambda c: c.ff,
        )
        self.assertFalse(
            f.filter_change(Change(project="x", repository="x", branch="x", category="x", ff=False))
        )
        self.assertFalse(
            f.filter_change(Change(project="p", repository="r", branch="b", category="c", ff=False))
        )
        self.assertFalse(
            f.filter_change(Change(project="x", repository="x", branch="x", category="x", ff=True))
        )
        self.assertTrue(
            f.filter_change(Change(project="p", repository="r", branch="b", category="c", ff=True))
        )

    def test_filter_props(self):
        f = ChangeFilter(property_eq={"event.type": "ref-updated"})
        self.assertTrue(f.filter_change(Change(properties={"event.type": "ref-updated"})))
        self.assertFalse(f.filter_change(Change(properties={"event.type": "patch-uploaded"})))
        self.assertFalse(f.filter_change(Change(properties={})))

        f = ChangeFilter(property_re={"event.type": "^ref-updated$"})
        self.assertTrue(f.filter_change(Change(properties={"event.type": "ref-updated"})))
        self.assertFalse(f.filter_change(Change(properties={"event.type": "patch-uploaded"})))
        self.assertFalse(f.filter_change(Change(properties={})))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import error
from twisted.python.reflect import namedModule
from twisted.trial import unittest

from buildbot import config
from buildbot.process import remotetransfer
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.steps.source import mercurial
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectDownloadFile
from buildbot.test.steps import ExpectRemoteRef
from buildbot.test.steps import ExpectRmdir
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import ExpectStat
from buildbot.test.util import sourcesteps


class TestMercurial(sourcesteps.SourceStepMixin, TestReactorMixin,
                    unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        return self.setUpSourceStep()

    def tearDown(self):
        return self.tearDownSourceStep()

    def patch_workerVersionIsOlderThan(self, result):
        self.patch(
            mercurial.Mercurial, 'workerVersionIsOlderThan', lambda x, y, z: result)

    def test_no_repourl(self):
        with self.assertRaises(config.ConfigErrors):
            mercurial.Mercurial(mode="full")

    def test_incorrect_mode(self):
        with self.assertRaises(config.ConfigErrors):
            mercurial.Mercurial(repourl='http://hg.mozilla.org', mode='invalid')

    def test_incorrect_method(self):
        with self.assertRaises(config.ConfigErrors):
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                method='invalid')

    def test_incorrect_branchType(self):
        with self.assertRaises(config.ConfigErrors):
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                branchType='invalid')

    def test_mode_full_clean(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='clean', branchType='inrepo'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--config',
                                 'extensions.purge=', 'purge'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_win32path(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='clean', branchType='inrepo'))
        self.build.path_module = namedModule('ntpath')
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file=r'wkdir\.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file=r'wkdir\.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--config',
                                 'extensions.purge=', 'purge'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_timeout(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                timeout=1,
                                mode='full', method='clean', branchType='inrepo'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        timeout=1,
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        timeout=1,
                        command=['hg', '--verbose', '--config',
                                 'extensions.purge=', 'purge'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        timeout=1,
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        timeout=1,
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        timeout=1,
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        timeout=1,
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        timeout=1,
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_patch(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='clean', branchType='inrepo'),
            patch=(1, 'patch'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(0),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--config',
                                 'extensions.purge=', 'purge'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'])
            .exit(0),
            ExpectDownloadFile(blocksize=32768, maxsize=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                               workerdest='.buildbot-diff', workdir='wkdir', mode=None)
            .exit(0),
            ExpectDownloadFile(blocksize=32768, maxsize=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                               workerdest='.buildbot-patched', workdir='wkdir', mode=None)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=[
                            'hg', '--verbose', 'import', '--no-commit', '-p', '1', '-'],
                        initial_stdin='patch')
            .exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_patch_worker_2_16(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='clean', branchType='inrepo'),
            patch=(1, 'patch'),
            worker_version={'*': '2.16'})
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(0),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--config',
                                 'extensions.purge=', 'purge'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'])
            .exit(0),
            ExpectDownloadFile(blocksize=32768, maxsize=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                               slavedest='.buildbot-diff', workdir='wkdir', mode=None)
            .exit(0),
            ExpectDownloadFile(blocksize=32768, maxsize=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                               slavedest='.buildbot-patched', workdir='wkdir', mode=None)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=[
                            'hg', '--verbose', 'import', '--no-commit', '-p', '1', '-'],
                        initial_stdin='patch')
            .exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_patch_fail(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='clean', branchType='inrepo'),
            patch=(1, 'patch'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(0),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--config',
                                 'extensions.purge=', 'purge'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'])
            .exit(0),
            ExpectDownloadFile(blocksize=32768, maxsize=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                               workerdest='.buildbot-diff', workdir='wkdir', mode=None)
            .exit(0),
            ExpectDownloadFile(blocksize=32768, maxsize=None,
                               reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                               workerdest='.buildbot-patched', workdir='wkdir', mode=None)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=[
                            'hg', '--verbose', 'import', '--no-commit', '-p', '1', '-'],
                        initial_stdin='patch')
            .exit(1)
        )
        self.expect_outcome(result=FAILURE, state_string="update (failure)")
        return self.run_step()

    def test_mode_full_clean_no_existing_repo(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='clean', branchType='inrepo'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org', '.'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'],
                        log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clobber(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='clobber', branchType='inrepo'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org', '.'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_fresh(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='fresh', branchType='inrepo'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--config',
                                 'extensions.purge=', 'purge', '--all'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_fresh_no_existing_repo(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='fresh', branchType='inrepo'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org', '.'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'],
                        log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_fresh_retry(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='fresh', branchType='inrepo',
                                retry=(0, 2)))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org', '.'])
            .exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org', '.'])
            .exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org', '.'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'],
                        log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_no_existing_repo_dirname(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='incremental', branchType='dirname'),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(1),  # does not exist
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org', '.'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update', '--clean'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_retry(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='incremental', branchType='dirname', retry=(0, 1)),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(1),  # does not exist
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org', '.'])
            .exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org', '.'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update', '--clean'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_branch_change_dirname(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org/',
                                mode='incremental', branchType='dirname', defaultBranch='devel'),
            dict(branch='stable')
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org/stable'])
            .exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org/stable', '.'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_no_existing_repo_inrepo(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='incremental', branchType='inrepo'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(1),  # does not exist
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org', '.'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update', '--clean',
                                 '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_existing_repo(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='incremental', branchType='inrepo'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),  # directory exists
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update', '--clean',
                                 '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_existing_repo_added_files(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='incremental', branchType='inrepo'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),  # directory exists
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .stdout('foo\nbar/baz\n')
            .exit(1),
            ExpectRmdir(dir=['wkdir/foo', 'wkdir/bar/baz'], log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update', '--clean',
                                 '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_existing_repo_added_files_old_rmdir(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='incremental', branchType='inrepo'))
        self.patch_workerVersionIsOlderThan(True)
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),  # directory exists
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .stdout('foo\nbar/baz\n')
            .exit(1),
            ExpectRmdir(dir='wkdir/foo', log_environ=True)
            .exit(0),
            ExpectRmdir(dir='wkdir/bar/baz', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update', '--clean',
                                 '--rev', 'default'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_given_revision(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='incremental', branchType='inrepo'), dict(
                revision='abcdef01',
            ))

        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'abcdef01'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update', '--clean',
                                 '--rev', 'abcdef01'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_branch_change(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='incremental', branchType='inrepo'), dict(
                branch='stable',
            ))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'stable'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'clone', '--noupdate',
                                 'http://hg.mozilla.org', '.'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'stable'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_branch_change_no_clobberOnBranchChange(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='incremental', branchType='inrepo',
                                clobberOnBranchChange=False), dict(
                branch='stable',
            ))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'stable'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'])
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'])
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'stable'])
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'])
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_env(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='clean', branchType='inrepo',
                                env={'abc': '123'}))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'], env={'abc': '123'})
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=True)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--config',
                                 'extensions.purge=', 'purge'], env={'abc': '123'})
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'], env={'abc': '123'})
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'],
                        env={'abc': '123'})
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'],
                        env={'abc': '123'})
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update', '--clean',
                                 '--rev', 'default'], env={'abc': '123'})
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'], env={'abc': '123'})
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_full_clean_log_environ(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='clean',
                                branchType='inrepo',
                                logEnviron=False))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'],
                        log_environ=False)
            .exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=False)
            .exit(1),
            ExpectStat(file='wkdir/.hg', log_environ=False)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--config',
                                 'extensions.purge=', 'purge'],
                        log_environ=False)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'pull',
                                 'http://hg.mozilla.org', '--rev', 'default'],
                        log_environ=False)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'identify', '--branch'],
                        log_environ=False)
            .stdout('default')
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'locate', 'set:added()'],
                        log_environ=False)
            .exit(1),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'update',
                                 '--clean', '--rev', 'default'],
                        log_environ=False)
            .exit(0),
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', 'parents',
                                 '--template', '{node}\\n'],
                        log_environ=False)
            .stdout('\n')
            .stdout('f6ad368298bd941e934a41f3babc827b2aa95a1d')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_command_fails(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='fresh', branchType='inrepo'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .exit(1)
        )
        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_worker_connection_lost(self):
        self.setup_step(
            mercurial.Mercurial(repourl='http://hg.mozilla.org',
                                mode='full', method='clean', branchType='inrepo'))
        self.expect_commands(
            ExpectShell(workdir='wkdir',
                        command=['hg', '--verbose', '--version'])
            .error(error.ConnectionLost())
        )
        self.expect_outcome(result=RETRY, state_string="update (retry)")
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from buildbot.config.errors import error


def check_param_length(value, name, max_length):
    if isinstance(value, str) and len(value) > max_length:
        error(f"{name} '{value}' exceeds maximum length of {max_length}")

    qualified_name = f"{type(value).__module__}.{type(value).__name__}"
    if qualified_name == 'buildbot.process.properties.Interpolate':
        if value.args:
            interpolations = tuple([''] * len(value.args))
        else:
            interpolations = {k: '' for k in value.interpolations}
        shortest_value = value.fmtstring % interpolations
        if len(shortest_value) > max_length:
            error(f"{name} '{value}' (shortest interpolation) exceeds maximum length of "
                  f"{max_length}")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import base64

from twisted.application import internet
from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet import task
from twisted.python import log
from twisted.words.protocols import irc

from buildbot import config
from buildbot.process.results import CANCELLED
from buildbot.process.results import EXCEPTION
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SKIPPED
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.reporters.words import Channel
from buildbot.reporters.words import Contact
from buildbot.reporters.words import StatusBot
from buildbot.reporters.words import ThrottledClientFactory
from buildbot.reporters.words import dangerousCommand
from buildbot.util import service
from buildbot.util import ssl


class UsageError(ValueError):

    # pylint: disable=useless-super-delegation
    def __init__(self, string="Invalid usage", *more):
        # This is not useless as we change the default value of an argument.
        # This bug is reported as "fixed" but apparently, it is not.
        # https://github.com/PyCQA/pylint/issues/1085
        # (Maybe there is a problem with builtin exceptions).
        super().__init__(string, *more)


_irc_colors = (
    'WHITE',
    'BLACK',
    'NAVY_BLUE',
    'GREEN',
    'RED',
    'BROWN',
    'PURPLE',
    'OLIVE',
    'YELLOW',
    'LIME_GREEN',
    'TEAL',
    'AQUA_LIGHT',
    'ROYAL_BLUE',
    'PINK',
    'DARK_GRAY',
    'LIGHT_GRAY'
)


class IRCChannel(Channel):

    def __init__(self, bot, channel):
        super().__init__(bot, channel)
        self.muted = False

    def send(self, message, **kwargs):
        if self.id[0] in irc.CHANNEL_PREFIXES:
            send = self.bot.groupSend
        else:
            send = self.bot.msg
        if not self.muted:
            send(self.id, message)

    def act(self, action):
        if self.muted:
            return
        self.bot.groupDescribe(self.id, action)


class IRCContact(Contact):

    def __init__(self, user, channel=None):
        if channel is None:
            channel = user
        super().__init__(user, channel)

    def act(self, action):
        return self.channel.act(action)

    def handleAction(self, action):
        # this is sent when somebody performs an action that mentions the
        # buildbot (like '/me kicks buildbot'). 'self.user' is the name/nick/id of
        # the person who performed the action, so if their action provokes a
        # response, they can be named.  This is 100% silly.
        if not action.endswith("s " + self.bot.nickname):
            return
        words = action.split()
        verb = words[-2]
        if verb == "kicks":
            response = f"{verb} back"
        elif verb == "threatens":
            response = f"hosts a red wedding for {self.user_id}"
        else:
            response = f"{verb} {self.user_id} too"
        self.act(response)

    @defer.inlineCallbacks
    def op_required(self, command):
        if self.is_private_chat or self.user_id in self.bot.authz.get(command.upper(), ()):
            return False
        ops = yield self.bot.getChannelOps(self.channel.id)
        return self.user_id not in ops

    # IRC only commands

    @dangerousCommand
    def command_JOIN(self, args, **kwargs):
        """join a channel"""
        args = self.splitArgs(args)
        for channel in args:
            self.bot.join(channel)
    command_JOIN.usage = "join #channel - join a channel #channel"

    @dangerousCommand
    def command_LEAVE(self, args, **kwargs):
        """join a channel"""
        args = self.splitArgs(args)
        for channel in args:
            self.bot.leave(channel)
    command_LEAVE.usage = "leave #channel - leave a channel #channel"

    @defer.inlineCallbacks
    def command_MUTE(self, args, **kwargs):
        if (yield self.op_required('mute')):
            yield self.send("Only channel operators or explicitly allowed users "
                            f"can mute me here, {self.user_id}... Blah, blah, blah...")
            return
        # The order of these is important! ;)
        yield self.send("Shutting up for now.")
        self.channel.muted = True
    command_MUTE.usage = "mute - suppress all messages until a corresponding 'unmute' is issued"

    @defer.inlineCallbacks
    def command_UNMUTE(self, args, **kwargs):
        if self.channel.muted:
            if (yield self.op_required('mute')):
                return
            # The order of these is important! ;)
            self.channel.muted = False
            yield self.send("I'm baaaaaaaaaaack!")
        else:
            yield self.send(
                "No one had told me to be quiet, but it's the thought that counts, right?")
    command_UNMUTE.usage = "unmute - disable a previous 'mute'"

    @defer.inlineCallbacks
    @Contact.overrideCommand
    def command_NOTIFY(self, args, **kwargs):
        if not self.is_private_chat:
            argv = self.splitArgs(args)
            if argv and argv[0] in ('on', 'off') and \
                    (yield self.op_required('notify')):
                yield self.send("Only channel operators can change notified events for this "
                                f"channel. And you, {self.user_id}, are neither!")
                return
        super().command_NOTIFY(args, **kwargs)

    def command_DANCE(self, args, **kwargs):
        """dance, dance academy..."""
        reactor.callLater(1.0, self.send, "<(^.^<)")
        reactor.callLater(2.0, self.send, "<(^.^)>")
        reactor.callLater(3.0, self.send, "(>^.^)>")
        reactor.callLater(3.5, self.send, "(7^.^)7")
        reactor.callLater(5.0, self.send, "(>^.^<)")

    def command_DESTROY(self, args):
        if self.bot.nickname not in args:
            self.act("readies phasers")
        else:
            self.send(f"Better destroy yourself, {self.user_id}!")

    def command_HUSTLE(self, args):
        self.act("does the hustle")
    command_HUSTLE.usage = "dondon on #qutebrowser: qutebrowser-bb needs to learn to do the hustle"


class IrcStatusBot(StatusBot, irc.IRCClient):

    """I represent the buildbot to an IRC server.
    """

    contactClass = IRCContact
    channelClass = IRCChannel

    def __init__(self, nickname, password, join_channels, pm_to_nicks,
                 noticeOnChannel, *args, useColors=False, useSASL=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.nickname = nickname
        self.join_channels = join_channels
        self.pm_to_nicks = pm_to_nicks
        self.password = password
        self.hasQuit = 0
        self.noticeOnChannel = noticeOnChannel
        self.useColors = useColors
        self.useSASL = useSASL
        self._keepAliveCall = task.LoopingCall(
            lambda: self.ping(self.nickname))
        self._channel_names = {}

    def register(self, nickname, hostname="foo", servername="bar"):
        if not self.useSASL:
            super().register(nickname, hostname, servername)
            return

        if self.password is not None:
            self.sendLine("CAP REQ :sasl")
        self.setNick(nickname)
        if self.username is None:
            self.username = nickname
        self.sendLine(
            f"USER {self.username} {hostname} {servername} :{self.realname}"
        )
        if self.password is not None:
            self.sendLine("AUTHENTICATE PLAIN")

    def irc_AUTHENTICATE(self, prefix, params):
        nick = self.nickname.encode()
        passwd = self.password.encode()
        code = base64.b64encode(nick + b'\0' + nick + b'\0' + passwd)
        self.sendLine("AUTHENTICATE " + code.decode())
        self.sendLine("CAP END")

    def connectionMade(self):
        super().connectionMade()
        self._keepAliveCall.start(60)

    def connectionLost(self, reason):
        if self._keepAliveCall.running:
            self._keepAliveCall.stop()
        super().connectionLost(reason)

    # The following methods are called when we write something.
    def groupSend(self, channel, message):
        if self.noticeOnChannel:
            self.notice(channel, message)
        else:
            self.msg(channel, message)

    def groupDescribe(self, channel, action):
        self.describe(channel, action)

    def getContact(self, user, channel=None):
        # nicknames and channel names are case insensitive
        user = user.lower()
        if channel is None:
            channel = user
        channel = channel.lower()
        return super().getContact(user, channel)

    # the following irc.IRCClient methods are called when we have input
    def privmsg(self, user, channel, message):
        user = user.split('!', 1)[0]  # rest is ~user@hostname
        # channel is '#twisted' or 'buildbot' (for private messages)
        if channel == self.nickname:
            # private message
            contact = self.getContact(user=user)
            d = contact.handleMessage(message)
            return d
        # else it's a broadcast message, maybe for us, maybe not. 'channel'
        # is '#twisted' or the like.
        contact = self.getContact(user=user, channel=channel)
        if message.startswith(f"{self.nickname}:") or message.startswith(f"{self.nickname},"):
            message = message[len(f"{self.nickname}:"):]
            d = contact.handleMessage(message)
            return d
        return None

    def action(self, user, channel, data):
        user = user.split('!', 1)[0]  # rest is ~user@hostname
        # somebody did an action (/me actions) in the broadcast channel
        contact = self.getContact(user=user, channel=channel)
        if self.nickname in data:
            contact.handleAction(data)

    def signedOn(self):
        if self.password:
            self.msg("Nickserv", "IDENTIFY " + self.password)
        for c in self.join_channels:
            if isinstance(c, dict):
                channel = c.get('channel', None)
                password = c.get('password', None)
            else:
                channel = c
                password = None
            self.join(channel=channel, key=password)
        for c in self.pm_to_nicks:
            contact = self.getContact(c)
            contact.channel.add_notification_events(self.notify_events)
        self.loadState()

    def getNames(self, channel):
        channel = channel.lower()
        d = defer.Deferred()
        callbacks = self._channel_names.setdefault(channel, ([], []))[0]
        callbacks.append(d)
        self.sendLine(f"NAMES {channel}")
        return d

    def irc_RPL_NAMREPLY(self, prefix, params):
        channel = params[2].lower()
        if channel not in self._channel_names:
            return
        nicks = params[3].split(' ')
        nicklist = self._channel_names[channel][1]
        nicklist += nicks

    def irc_RPL_ENDOFNAMES(self, prefix, params):
        channel = params[1].lower()
        try:
            callbacks, namelist = self._channel_names.pop(channel)
        except KeyError:
            return
        for cb in callbacks:
            cb.callback(namelist)

    @defer.inlineCallbacks
    def getChannelOps(self, channel):
        names = yield self.getNames(channel)
        return [n[1:] for n in names if n[0] in '@&~%']

    def joined(self, channel):
        self.log(f"Joined {channel}")
        # trigger contact constructor, which in turn subscribes to notify events
        channel = self.getChannel(channel=channel)
        channel.add_notification_events(self.notify_events)

    def left(self, channel):
        self.log(f"Left {channel}")

    def kickedFrom(self, channel, kicker, message):
        self.log(f"I have been kicked from {channel} by {kicker}: {message}")

    def userLeft(self, user, channel):
        if user:
            user = user.lower()
        if channel:
            channel = channel.lower()
        if (channel, user) in self.contacts:
            del self.contacts[(channel, user)]

    def userKicked(self, kickee, channel, kicker, message):
        self.userLeft(kickee, channel)

    def userQuit(self, user, quitMessage=None):
        if user:
            user = user.lower()
        for c, u in list(self.contacts):
            if u == user:
                del self.contacts[(c, u)]

    results_colors = {
        SUCCESS: 'GREEN',
        WARNINGS: 'YELLOW',
        FAILURE: 'RED',
        SKIPPED: 'ROYAL_BLUE',
        EXCEPTION: 'PURPLE',
        RETRY: 'AQUA_LIGHT',
        CANCELLED: 'PINK',
    }

    short_results_descriptions = {
        SUCCESS: ", Success",
        WARNINGS: ", Warnings",
        FAILURE: ", Failure",
        SKIPPED: ", Skipped",
        EXCEPTION: ", Exception",
        RETRY: ", Retry",
        CANCELLED: ", Cancelled",
    }

    def format_build_status(self, build, short=False):
        br = build['results']
        if short:
            text = self.short_results_descriptions[br]
        else:
            text = self.results_descriptions[br]
        if self.useColors:
            return f"\x03{_irc_colors.index(self.results_colors[br])}{text}\x0f"
        else:
            return text


class IrcStatusFactory(ThrottledClientFactory):
    protocol = IrcStatusBot

    shuttingDown = False
    p = None

    def __init__(self, nickname, password, join_channels, pm_to_nicks, authz, tags, notify_events,
                 noticeOnChannel=False,
                 useRevisions=False, showBlameList=False,
                 useSASL=False,
                 parent=None,
                 lostDelay=None, failedDelay=None, useColors=True):
        super().__init__(lostDelay=lostDelay, failedDelay=failedDelay)
        self.nickname = nickname
        self.password = password
        self.join_channels = join_channels
        self.pm_to_nicks = pm_to_nicks
        self.tags = tags
        self.authz = authz
        self.parent = parent
        self.notify_events = notify_events
        self.noticeOnChannel = noticeOnChannel
        self.useRevisions = useRevisions
        self.showBlameList = showBlameList
        self.useColors = useColors
        self.useSASL = useSASL

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['p']
        return d

    def shutdown(self):
        self.shuttingDown = True
        if self.p:
            self.p.quit("buildmaster reconfigured: bot disconnecting")

    def buildProtocol(self, address):
        if self.p:
            self.p.disownServiceParent()

        p = self.protocol(self.nickname, self.password,
                          self.join_channels, self.pm_to_nicks,
                          self.noticeOnChannel, self.authz,
                          self.tags, self.notify_events,
                          useColors=self.useColors,
                          useSASL=self.useSASL,
                          useRevisions=self.useRevisions,
                          showBlameList=self.showBlameList)
        p.setServiceParent(self.parent)
        p.factory = self
        self.p = p
        return p

    # TODO: I think a shutdown that occurs while the connection is being
    # established will make this explode

    def clientConnectionLost(self, connector, reason):
        if self.shuttingDown:
            log.msg("not scheduling reconnection attempt")
            return
        super().clientConnectionLost(connector, reason)

    def clientConnectionFailed(self, connector, reason):
        if self.shuttingDown:
            log.msg("not scheduling reconnection attempt")
            return
        super().clientConnectionFailed(connector, reason)


class IRC(service.BuildbotService):
    name = "IRC"
    in_test_harness = False
    f = None
    compare_attrs = ("host", "port", "nick", "password", "authz",
                     "channels", "pm_to_nicks", "useSSL",
                     "useSASL",
                     "useRevisions", "tags", "useColors",
                     "allowForce", "allowShutdown",
                     "lostDelay", "failedDelay")
    secrets = ['password']

    def checkConfig(self, host, nick, channels, pm_to_nicks=None, port=6667,
                    allowForce=None, tags=None, password=None, notify_events=None,
                    showBlameList=True, useRevisions=False, useSSL=False,
                    useSASL=False, lostDelay=None, failedDelay=None, useColors=True,
                    allowShutdown=None, noticeOnChannel=False, authz=None, **kwargs
                    ):
        deprecated_params = list(kwargs)
        if deprecated_params:
            config.error(f'{",".join(deprecated_params)} are deprecated')

        # deprecated
        if allowForce is not None:
            if authz is not None:
                config.error("If you specify authz, you must not use allowForce anymore")
            if allowForce not in (True, False):
                config.error(f"allowForce must be boolean, not {repr(allowForce)}")
            log.msg('IRC: allowForce is deprecated: use authz instead')
        if allowShutdown is not None:
            if authz is not None:
                config.error("If you specify authz, you must not use allowShutdown anymore")
            if allowShutdown not in (True, False):
                config.error(f"allowShutdown must be boolean, not {repr(allowShutdown)}")
            log.msg('IRC: allowShutdown is deprecated: use authz instead')
        # ###

        if noticeOnChannel not in (True, False):
            config.error(f"noticeOnChannel must be boolean, not {repr(noticeOnChannel)}")
        if useSSL:
            # SSL client needs a ClientContextFactory for some SSL mumbo-jumbo
            ssl.ensureHasSSL(self.__class__.__name__)
        if authz is not None:
            for acl in authz.values():
                if not isinstance(acl, (list, tuple, bool)):
                    config.error(
                        "authz values must be bool or a list of nicks")

    def reconfigService(self, host, nick, channels, pm_to_nicks=None, port=6667,
                        allowForce=None, tags=None, password=None, notify_events=None,
                        showBlameList=True, useRevisions=False, useSSL=False,
                        useSASL=False, lostDelay=None, failedDelay=None, useColors=True,
                        allowShutdown=None, noticeOnChannel=False, authz=None, **kwargs
                        ):

        # need to stash these so we can detect changes later
        self.host = host
        self.port = port
        self.nick = nick
        self.join_channels = channels
        if pm_to_nicks is None:
            pm_to_nicks = []
        self.pm_to_nicks = pm_to_nicks
        self.password = password
        if authz is None:
            self.authz = {}
        else:
            self.authz = authz
        self.useRevisions = useRevisions
        self.tags = tags
        if notify_events is None:
            notify_events = {}
        self.notify_events = notify_events
        self.noticeOnChannel = noticeOnChannel

        # deprecated...
        if allowForce is not None:
            self.authz[('force', 'stop')] = allowForce
        if allowShutdown is not None:
            self.authz[('shutdown')] = allowShutdown
        # ###

        # This function is only called in case of reconfig with changes
        # We don't try to be smart here. Just restart the bot if config has
        # changed.
        if self.f is not None:
            self.f.shutdown()
        self.f = IrcStatusFactory(self.nick, self.password,
                                  self.join_channels, self.pm_to_nicks,
                                  self.authz, self.tags,
                                  self.notify_events, parent=self,
                                  noticeOnChannel=noticeOnChannel,
                                  useRevisions=useRevisions,
                                  useSASL=useSASL,
                                  showBlameList=showBlameList,
                                  lostDelay=lostDelay,
                                  failedDelay=failedDelay,
                                  useColors=useColors)

        if useSSL:
            cf = ssl.ClientContextFactory()
            c = internet.SSLClient(self.host, self.port, self.f, cf)
        else:
            c = internet.TCPClient(self.host, self.port, self.f)

        c.setServiceParent(self)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from buildbot.steps.source.git import Git


class Gerrit(Git):

    def run_vc(self, branch, revision, patch):
        gerrit_branch = None

        changed_project = self.build.getProperty('event.change.project')
        if (not self.sourcestamp or (self.sourcestamp.project !=
                                     changed_project)):
            # If we don't have a sourcestamp, or the project is  wrong, this
            # isn't the repo that's changed.  Drop through and check out the
            # head of the given branch
            pass
        elif self.build.hasProperty("event.patchSet.ref"):
            gerrit_branch = self.build.getProperty("event.patchSet.ref")
            self.updateSourceProperty("gerrit_branch", gerrit_branch)
        else:
            try:
                change = self.build.getProperty("gerrit_change", '').split('/')
                if len(change) == 2:
                    gerrit_branch = \
                        f"refs/changes/{(int(change[0]) % 100):2}/{int(change[0])}/{int(change[1])}"
                    self.updateSourceProperty("gerrit_branch", gerrit_branch)
            except Exception:
                pass

        branch = gerrit_branch or branch
        return super().run_vc(branch, revision, patch)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import json

from twisted.internet import defer
from twisted.python import log
from twisted.web.error import Error

from buildbot.util import bytes2unicode
from buildbot.util import unicode2bytes
from buildbot.www import resource
from buildbot.www.rest import RestRootResource


class V3RootResource(resource.Resource):
    isLeaf = True

    # enable reconfigResource calls
    needsReconfig = True

    def reconfigResource(self, new_config):
        # @todo v2 has cross origin support, which might need to be factorized
        graphql_config = new_config.www.get("graphql")
        self.debug = True
        self.graphql = None
        if graphql_config is not None:
            self.graphql = True

    def render(self, request):
        def writeError(msg, errcode=400):
            if isinstance(msg, list):
                errors = msg
            else:
                msg = bytes2unicode(msg)
                errors = [{"message": msg}]
            if self.debug:
                log.msg(f"HTTP error: {errors}")
            request.setResponseCode(errcode)
            request.setHeader(b"content-type", b"application/json; charset=utf-8")
            data = json.dumps({"data": None, "errors": errors})
            data = unicode2bytes(data)
            request.write(data)
            request.finish()

        return self.asyncRenderHelper(request, self.asyncRender, writeError)

    @defer.inlineCallbacks
    def asyncRender(self, request):
        if self.graphql is None:
            raise Error(501, "graphql not enabled")

        # graphql accepts its query either in post data or get query
        if request.method == b"POST":
            content_type = request.getHeader(b"content-type")
            if content_type == b"application/graphql":
                query = request.content.read().decode()
            elif content_type == b"application/json":
                json_query = json.load(request.content)
                query = json_query.pop('query')
                if json_query:
                    fields = " ".join(json_query.keys())
                    raise Error(400, b"json request unsupported fields: " + fields.encode())
            elif content_type is None:
                raise Error(400, b"no content-type")
            else:
                raise Error(400, b"unsupported content-type: " + content_type)

        elif request.method in (b"GET"):
            if b"query" not in request.args:
                raise Error(400, b"GET request must contain a 'query' parameter")
            query = request.args[b"query"][0].decode()
        else:
            raise Error(400, b"invalid HTTP method")

        res = yield self.master.graphql.query(query)
        errors = None
        if res.errors:
            errors = [e.formatted for e in res.errors]
        request.setHeader(b"content-type", b"application/json; charset=utf-8")
        data = json.dumps({"data": res.data, "errors": errors}).encode()
        request.write(data)


RestRootResource.addApiVersion(3, V3RootResource)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.util.protocol import LineProcessProtocol


class FakeLineProcessProtocol(LineProcessProtocol):

    def __init__(self):
        super().__init__()
        self.out_lines = []
        self.err_lines = []

    def outLineReceived(self, line):
        self.out_lines.append(line)

    def errLineReceived(self, line):
        self.err_lines.append(line)


class TestLineProcessProtocol(unittest.TestCase):

    def test_stdout(self):
        p = FakeLineProcessProtocol()
        p.outReceived(b'\nline2\nline3\nli')
        p.outReceived(b'ne4\nli')
        self.assertEqual(p.out_lines, [b'', b'line2', b'line3', b'line4'])
        p.processEnded(0)
        self.assertEqual(p.out_lines, [b'', b'line2', b'line3', b'line4', b'li'])

    def test_stderr(self):
        p = FakeLineProcessProtocol()
        p.errReceived(b'\nline2\nline3\nli')
        p.errReceived(b'ne4\nli')
        self.assertEqual(p.err_lines, [b'', b'line2', b'line3', b'line4'])
        p.processEnded(0)
        self.assertEqual(p.err_lines, [b'', b'line2', b'line3', b'line4', b'li'])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import pprint

from twisted.internet import defer
from twisted.python import log

from buildbot.mq import base
from buildbot.util import service
from buildbot.util import tuplematch


class SimpleMQ(service.ReconfigurableServiceMixin, base.MQBase):

    def __init__(self):
        super().__init__()
        self.qrefs = []
        self.persistent_qrefs = {}
        self.debug = False

    def reconfigServiceWithBuildbotConfig(self, new_config):
        self.debug = new_config.mq.get('debug', False)
        return super().reconfigServiceWithBuildbotConfig(new_config)

    def produce(self, routingKey, data):
        if self.debug:
            log.msg(f"MSG: {routingKey}\n{pprint.pformat(data)}")
        for qref in self.qrefs:
            if tuplematch.matchTuple(routingKey, qref.filter):
                self.invokeQref(qref, routingKey, data)

    def startConsuming(self, callback, filter, persistent_name=None):
        if any(not isinstance(k, str) and k is not None for k in filter):
            raise AssertionError(f"{filter} is not a filter")
        if persistent_name:
            if persistent_name in self.persistent_qrefs:
                qref = self.persistent_qrefs[persistent_name]
                qref.startConsuming(callback)
            else:
                qref = PersistentQueueRef(self, callback, filter)
                self.qrefs.append(qref)
                self.persistent_qrefs[persistent_name] = qref
        else:
            qref = QueueRef(self, callback, filter)
            self.qrefs.append(qref)
        return defer.succeed(qref)


class QueueRef(base.QueueRef):

    __slots__ = ['mq', 'filter']

    def __init__(self, mq, callback, filter):
        super().__init__(callback)
        self.mq = mq
        self.filter = filter

    def stopConsuming(self):
        self.callback = None
        try:
            self.mq.qrefs.remove(self)
        except ValueError:
            pass


class PersistentQueueRef(QueueRef):

    __slots__ = ['active', 'queue']

    def __init__(self, mq, callback, filter):
        super().__init__(mq, callback, filter)
        self.queue = []

    def startConsuming(self, callback):
        self.callback = callback
        self.active = True

        # invoke for every message that was missed
        queue, self.queue = self.queue, []
        for routingKey, data in queue:
            self.invoke(routingKey, data)

    def stopConsuming(self):
        self.callback = self.addToQueue
        self.active = False

    def addToQueue(self, routingKey, data):
        self.queue.append((routingKey, data))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.python import log

from buildbot.statistics.storage_backends.base import StatsStorageBase
from buildbot.util import service


class StatsService(service.BuildbotService):

    """
    A middleware for passing on statistics data to all storage backends.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.consumers = []

    def checkConfig(self, storage_backends):
        for wfb in storage_backends:
            if not isinstance(wfb, StatsStorageBase):
                raise TypeError(f"Invalid type of stats storage service {type(StatsStorageBase)!r}."
                                " Should be of type StatsStorageBase, "
                                f"is: {type(StatsStorageBase)!r}")

    @defer.inlineCallbacks
    def reconfigService(self, storage_backends):
        log.msg(f"Reconfiguring StatsService with config: {storage_backends!r}")

        self.checkConfig(storage_backends)

        self.registeredStorageServices = []
        for svc in storage_backends:
            self.registeredStorageServices.append(svc)

        yield self.removeConsumers()
        yield self.registerConsumers()

    @defer.inlineCallbacks
    def registerConsumers(self):
        self.consumers = []

        for svc in self.registeredStorageServices:
            for cap in svc.captures:
                cap.parent_svcs.append(svc)
                cap.master = self.master
                consumer = yield self.master.mq.startConsuming(cap.consume, cap.routingKey)
                self.consumers.append(consumer)

    @defer.inlineCallbacks
    def stopService(self):
        yield super().stopService()
        yield self.removeConsumers()

    @defer.inlineCallbacks
    def removeConsumers(self):
        for consumer in self.consumers:
            yield consumer.stopConsuming()
        self.consumers = []

    @defer.inlineCallbacks
    def yieldMetricsValue(self, data_name, post_data, buildid):
        """
        A method to allow posting data that is not generated and stored as build-data in
        the database. This method generates the `stats-yield-data` event to the mq layer
        which is then consumed in self.postData.

        @params
        data_name: (str) The unique name for identifying this data.
        post_data: (dict) A dictionary of key-value pairs that'll be sent for storage.
        buildid: The buildid of the current Build.
        """
        build_data = yield self.master.data.get(('builds', buildid))
        routingKey = ("stats-yieldMetricsValue", "stats-yield-data")

        msg = {
            'data_name': data_name,
            'post_data': post_data,
            'build_data': build_data
        }

        self.master.mq.produce(routingKey, msg)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import socket

from parameterized import parameterized

import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot import config
from buildbot.interfaces import LatentWorkerFailedToSubstantiate
from buildbot.test.fake import libvirt as libvirtfake
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.runprocess import ExpectMasterShell
from buildbot.test.runprocess import MasterRunProcessMixin
from buildbot.test.util.warnings import assertProducesWarnings
from buildbot.warnings import DeprecatedApiWarning
from buildbot.worker import libvirt as libvirtworker


# The libvirt module has a singleton threadpool within the module which we can't use in tests as
# this makes it impossible to run them concurrently. To work around this we introduce a per-test
# threadpool and access it through a class instance
class TestThreadWithQueue(libvirtworker.ThreadWithQueue):
    def __init__(self, pool, uri):
        super().__init__(pool, uri, connect_backoff_start_seconds=0, connect_backoff_multiplier=0,
                         connect_backoff_max_wait_seconds=0)

    def libvirt_open(self):
        return self.pool.case.libvirt_open(self.uri)


class TestServerThreadPool(libvirtworker.ServerThreadPool):
    ThreadClass = TestThreadWithQueue

    def __init__(self, case):
        super().__init__()
        self.case = case


class TestLibvirtWorker(libvirtworker.LibVirtWorker):
    def __init__(self, case, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.case = case
        self.pool = case.threadpool


class TestException(Exception):
    pass


class TestLibVirtWorker(TestReactorMixin, MasterRunProcessMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        self.setup_master_run_process()
        self.connections = {}
        self.patch(libvirtworker, "libvirt", libvirtfake)
        self.threadpool = TestServerThreadPool(self)

    def libvirt_open(self, uri):
        if uri not in self.connections:
            raise Exception('Could not find test connection')
        return self.connections[uri]

    def add_fake_conn(self, uri):
        conn = libvirtfake.Connection(uri)
        self.connections[uri] = conn
        return conn

    def create_worker(self, *args, **kwargs):
        worker = TestLibvirtWorker(self, *args, **kwargs)
        worker.parent = mock.Mock()
        worker.parent.master = mock.Mock()
        worker.parent.master.reactor = self.reactor
        return worker

    def raise_libvirt_error(self):
        # Helper method to be used from lambdas as they don't accept statements
        raise libvirtfake.libvirtError()

    def test_constructor_nolibvirt(self):
        self.patch(libvirtworker, "libvirt", None)
        with self.assertRaises(config.ConfigErrors):
            self.create_worker('bot', 'pass', None, 'path', 'path')

    def test_deprecated_connection(self):
        with assertProducesWarnings(DeprecatedApiWarning,
                                    message_pattern='connection argument has been deprecated'):
            self.create_worker('bot', 'pass', libvirtworker.Connection('test'), 'path', 'path')

    def test_deprecated_connection_and_uri(self):
        with self.assertRaises(config.ConfigErrors):
            with assertProducesWarnings(DeprecatedApiWarning,
                                        message_pattern='connection argument has been deprecated'):
                self.create_worker('bot', 'pass', libvirtworker.Connection('test'), 'path', 'path',
                                   uri='custom')

    @defer.inlineCallbacks
    def test_get_domain_id(self):
        conn = self.add_fake_conn('fake:///conn')
        conn.fake_add('bot', 14)

        bs = self.create_worker('bot', 'pass', hd_image='p', base_image='o', uri='fake:///conn')

        id = yield bs._get_domain_id()
        self.assertEqual(id, 14)

    @defer.inlineCallbacks
    def test_prepare_base_image_none(self):
        bs = self.create_worker('bot', 'pass', hd_image='p', base_image=None)
        yield bs._prepare_base_image()

        self.assert_all_commands_ran()

    @defer.inlineCallbacks
    def test_prepare_base_image_cheap(self):
        self.expect_commands(
            ExpectMasterShell(["qemu-img", "create", "-b", "o", "-f", "qcow2", "p"])
        )

        bs = self.create_worker('bot', 'pass', hd_image='p', base_image='o')
        yield bs._prepare_base_image()

        self.assert_all_commands_ran()

    @defer.inlineCallbacks
    def test_prepare_base_image_full(self):
        self.expect_commands(
            ExpectMasterShell(["cp", "o", "p"])
        )

        bs = self.create_worker('bot', 'pass', hd_image='p', base_image='o')
        bs.cheap_copy = False
        yield bs._prepare_base_image()

        self.assert_all_commands_ran()

    @defer.inlineCallbacks
    def test_prepare_base_image_fail(self):
        self.expect_commands(
            ExpectMasterShell(["cp", "o", "p"])
            .exit(1)
        )

        bs = self.create_worker('bot', 'pass', hd_image='p', base_image='o')
        bs.cheap_copy = False
        with self.assertRaises(LatentWorkerFailedToSubstantiate):
            yield bs._prepare_base_image()

        self.assert_all_commands_ran()

    @defer.inlineCallbacks
    def _test_stop_instance(self, graceful, fast, expected_destroy,
                            expected_shutdown, shutdown_side_effect=None):
        domain = mock.Mock()
        domain.ID.side_effect = lambda: 14
        domain.shutdown.side_effect = shutdown_side_effect

        conn = self.add_fake_conn('fake:///conn')
        conn.fake_add_domain('name', domain)

        bs = self.create_worker('name', 'p', hd_image='p', base_image='o',
                                uri='fake:///conn', xml='<xml/>')
        bs.graceful_shutdown = graceful
        with mock.patch('os.remove') as remove_mock:
            yield bs.stop_instance(fast=fast)

        self.assertEqual(int(expected_destroy), domain.destroy.call_count)
        self.assertEqual(int(expected_shutdown), domain.shutdown.call_count)
        remove_mock.assert_called_once_with('p')

        self.assert_all_commands_ran()

    @defer.inlineCallbacks
    def test_stop_instance_destroy(self):
        yield self._test_stop_instance(graceful=False,
                                       fast=False,
                                       expected_destroy=True,
                                       expected_shutdown=False)

    @defer.inlineCallbacks
    def test_stop_instance_shutdown(self):
        yield self._test_stop_instance(graceful=True,
                                       fast=False,
                                       expected_destroy=False,
                                       expected_shutdown=True)

    @defer.inlineCallbacks
    def test_stop_instance_shutdown_fails(self):
        yield self._test_stop_instance(graceful=True,
                                       fast=False,
                                       expected_destroy=True,
                                       expected_shutdown=True,
                                       shutdown_side_effect=TestException)

    @defer.inlineCallbacks
    def test_start_instance_connection_fails(self):
        bs = self.create_worker('b', 'p', hd_image='p', base_image='o', uri='unknown')

        prep = mock.Mock()
        prep.side_effect = lambda: defer.succeed(0)
        self.patch(bs, "_prepare_base_image", prep)

        with self.assertRaisesRegex(LatentWorkerFailedToSubstantiate, 'Did not receive connection'):
            yield bs.start_instance(mock.Mock())

        self.assertFalse(prep.called)

    @defer.inlineCallbacks
    def test_start_instance_already_active(self):
        conn = self.add_fake_conn('fake:///conn')
        conn.fake_add('bot', 14)

        bs = self.create_worker('bot', 'p', hd_image='p', base_image='o', uri='fake:///conn',
                                xml='<xml/>')

        prep = mock.Mock()
        self.patch(bs, "_prepare_base_image", prep)

        with self.assertRaisesRegex(LatentWorkerFailedToSubstantiate, 'it\'s already active'):
            yield bs.start_instance(mock.Mock())

        self.assertFalse(prep.called)

    @defer.inlineCallbacks
    def test_start_instance_domain_id_error(self):
        conn = self.add_fake_conn('fake:///conn')
        domain = conn.fake_add('bot', 14)
        domain.ID = self.raise_libvirt_error

        bs = self.create_worker('bot', 'p', hd_image='p', base_image='o', uri='fake:///conn',
                                xml='<xml/>')

        prep = mock.Mock()
        self.patch(bs, "_prepare_base_image", prep)

        with self.assertRaisesRegex(LatentWorkerFailedToSubstantiate, 'while retrieving domain ID'):
            yield bs.start_instance(mock.Mock())

        self.assertFalse(prep.called)

    @defer.inlineCallbacks
    def test_start_instance_connection_create_fails(self):
        bs = self.create_worker('bot', 'p', hd_image='p', base_image='o', xml='<xml/>',
                                uri='fake:///conn')

        conn = self.add_fake_conn('fake:///conn')
        conn.createXML = lambda _, __: self.raise_libvirt_error()

        prep = mock.Mock()
        prep.side_effect = lambda: defer.succeed(0)
        self.patch(bs, "_prepare_base_image", prep)

        with self.assertRaisesRegex(LatentWorkerFailedToSubstantiate, 'error while starting VM'):
            yield bs.start_instance(mock.Mock())

        self.assertTrue(prep.called)

    @defer.inlineCallbacks
    def test_start_instance_domain_create_fails(self):
        bs = self.create_worker('bot', 'p', hd_image='p', base_image='o', uri='fake:///conn')

        conn = self.add_fake_conn('fake:///conn')
        domain = conn.fake_add('bot', -1)
        domain.create = self.raise_libvirt_error

        prep = mock.Mock()
        prep.side_effect = lambda: defer.succeed(0)
        self.patch(bs, "_prepare_base_image", prep)

        with self.assertRaisesRegex(LatentWorkerFailedToSubstantiate, 'error while starting VM'):
            yield bs.start_instance(mock.Mock())

        self.assertTrue(prep.called)

    @defer.inlineCallbacks
    def test_start_instance_xml(self):
        self.add_fake_conn('fake:///conn')

        bs = self.create_worker('bot', 'p', hd_image='p', base_image='o', uri='fake:///conn',
                                xml='<xml/>')

        prep = mock.Mock()
        prep.side_effect = lambda: defer.succeed(0)
        self.patch(bs, "_prepare_base_image", prep)

        started = yield bs.start_instance(mock.Mock())

        self.assertEqual(started, True)

    @parameterized.expand([
        ('set_fqdn', {'masterFQDN': 'somefqdn'}, 'somefqdn'),
        ('auto_fqdn', {}, socket.getfqdn()),
    ])
    @defer.inlineCallbacks
    def test_start_instance_existing_domain(self, name, kwargs, expect_fqdn):
        conn = self.add_fake_conn('fake:///conn')
        domain = conn.fake_add('bot', -1)

        bs = self.create_worker('bot', 'p', hd_image='p', base_image='o', uri='fake:///conn',
                                **kwargs)

        prep = mock.Mock()
        prep.side_effect = lambda: defer.succeed(0)
        self.patch(bs, "_prepare_base_image", prep)

        started = yield bs.start_instance(mock.Mock())

        self.assertEqual(started, True)
        self.assertEqual(domain.metadata, {
            'buildbot': (libvirtfake.VIR_DOMAIN_METADATA_ELEMENT,
                         'http://buildbot.net/',
                         f'<auth username="bot" password="p" master="{expect_fqdn}"/>',
                         libvirtfake.VIR_DOMAIN_AFFECT_CONFIG)
        })



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.data import properties
from buildbot.process.properties import Properties as processProperties
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import endpoint
from buildbot.test.util import interfaces


class BuildsetPropertiesEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = properties.BuildsetPropertiesEndpoint
    resourceTypeClass = properties.Properties

    def setUp(self):
        self.setUpEndpoint()
        self.db.insertTestData([
            fakedb.Buildset(id=13, reason='because I said so'),
            fakedb.SourceStamp(id=92),
            fakedb.SourceStamp(id=93),
            fakedb.BuildsetSourceStamp(buildsetid=13, sourcestampid=92),
            fakedb.BuildsetSourceStamp(buildsetid=13, sourcestampid=93),

            fakedb.Buildset(id=14, reason='no sourcestamps'),

            fakedb.BuildsetProperty(buildsetid=14)
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @defer.inlineCallbacks
    def test_get_properties(self):
        props = yield self.callGet(('buildsets', 14, 'properties'))

        self.assertEqual(props, {'prop': (22, 'fakedb')})


class BuildPropertiesEndpoint(endpoint.EndpointMixin, unittest.TestCase):

    endpointClass = properties.BuildPropertiesEndpoint
    resourceTypeClass = properties.Properties

    def setUp(self):
        self.setUpEndpoint()
        self.db.insertTestData([
            fakedb.Buildset(id=28),
            fakedb.BuildRequest(id=5, buildsetid=28),
            fakedb.Master(id=3),
            fakedb.Worker(id=42, name="Friday"),
            fakedb.Build(id=786, buildrequestid=5, masterid=3, workerid=42),
            fakedb.BuildProperty(
                buildid=786, name="year", value=1651, source="Wikipedia"),
            fakedb.BuildProperty(
                buildid=786, name="island_name", value="despair", source="Book"),
        ])

    def tearDown(self):
        self.tearDownEndpoint()

    @defer.inlineCallbacks
    def test_get_properties(self):
        props = yield self.callGet(('builds', 786, 'properties'))

        self.assertEqual(props, {'year': (1651, 'Wikipedia'),
                         'island_name': ("despair", 'Book')})

    @defer.inlineCallbacks
    def test_get_properties_from_builder(self):
        props = yield self.callGet(('builders', 1, 'builds', 786, 'properties'))
        self.assertEqual(props, {'year': (1651, 'Wikipedia'),
                         'island_name': ("despair", 'Book')})


class Properties(interfaces.InterfaceTests, TestReactorMixin,
                 unittest.TestCase):

    def setUp(self):
        self.setup_test_reactor()
        self.master = fakemaster.make_master(self, wantMq=False, wantDb=True,
                                             wantData=True)
        self.rtype = properties.Properties(self.master)

    @defer.inlineCallbacks
    def do_test_callthrough(self, dbMethodName, method, exp_args=None,
                            exp_kwargs=None, *args, **kwargs):
        rv = (1, 2)
        m = mock.Mock(return_value=defer.succeed(rv))
        setattr(self.master.db.builds, dbMethodName, m)
        res = yield method(*args, **kwargs)
        self.assertIdentical(res, rv)
        m.assert_called_with(
            *(exp_args or args), **((exp_kwargs is None) and kwargs or exp_kwargs))

    def test_signature_setBuildProperty(self):
        @self.assertArgSpecMatches(
            self.master.data.updates.setBuildProperty,  # fake
            self.rtype.setBuildProperty)  # real
        def setBuildProperty(self, buildid, name, value, source):
            pass

    def test_setBuildProperty(self):
        return self.do_test_callthrough('setBuildProperty', self.rtype.setBuildProperty,
                                        buildid=1234, name='property', value=[42, 45],
                                        source='testsuite',
                                        exp_args=(1234, 'property', [42, 45], 'testsuite'),
                                        exp_kwargs={})

    @defer.inlineCallbacks
    def test_setBuildProperties(self):
        self.master.db.insertTestData([
            fakedb.Buildset(id=28),
            fakedb.BuildRequest(id=5, buildsetid=28),
            fakedb.Master(id=3),
            fakedb.Worker(id=42, name="Friday"),
            fakedb.Build(id=1234, buildrequestid=5, masterid=3, workerid=42),
        ])

        self.master.db.builds.setBuildProperty = mock.Mock(
            wraps=self.master.db.builds.setBuildProperty)
        props = processProperties.fromDict(
            dict(a=(1, 't'), b=(['abc', 9], 't')))
        yield self.rtype.setBuildProperties(1234, props)
        setBuildPropertiesCalls = sorted(self.master.db.builds.setBuildProperty.mock_calls)
        self.assertEqual(setBuildPropertiesCalls, [
            mock.call(1234, 'a', 1, 't'),
            mock.call(1234, 'b', ['abc', 9], 't')])
        self.master.mq.assertProductions([
            (('builds', '1234', 'properties', 'update'),
             {'a': (1, 't'), 'b': (['abc', 9], 't')}),
        ])
        # sync without changes: no db write
        self.master.db.builds.setBuildProperty.reset_mock()
        self.master.mq.clearProductions()
        yield self.rtype.setBuildProperties(1234, props)
        self.master.db.builds.setBuildProperty.assert_not_called()
        self.master.mq.assertProductions([])

        # sync with one changes: one db write
        props.setProperty('b', 2, 'step')
        self.master.db.builds.setBuildProperty.reset_mock()
        yield self.rtype.setBuildProperties(1234, props)

        self.master.db.builds.setBuildProperty.assert_called_with(
            1234, 'b', 2, 'step')
        self.master.mq.assertProductions([
            (('builds', '1234', 'properties', 'update'), {'b': (2, 'step')})
        ])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from io import StringIO

from twisted.internet import defer

from buildbot.test.util.integration import RunMasterBase

# This integration test creates a master and worker environment,
# with two builders and a trigger step linking them

expectedOutputRegex = \
    r"""\*\*\* BUILD 1 \*\*\* ==> build successful \(success\)
    \*\*\* STEP worker_preparation \*\*\* ==> worker local1 ready \(success\)
    \*\*\* STEP shell \*\*\* ==> 'echo hello' \(success\)
        log:stdio \({loglines}\)
    \*\*\* STEP trigger \*\*\* ==> triggered trigsched, 1 success \(success\)
       url:trigsched #2 \(http://localhost:8080/#buildrequests/2\)
       url:success: build #1 \(http://localhost:8080/#builders/(1|2)/builds/1\)
    \*\*\* STEP shell_1 \*\*\* ==> 'echo world' \(success\)
        log:stdio \({loglines}\)
\*\*\* BUILD 2 \*\*\* ==> build successful \(success\)
    \*\*\* STEP worker_preparation \*\*\* ==> worker local1 ready \(success\)
    \*\*\* STEP shell \*\*\* ==> 'echo ola' \(success\)
        log:stdio \({loglines}\)
"""


class TriggeringMaster(RunMasterBase):

    change = dict(branch="master",
                  files=["foo.c"],
                  author="me@foo.com",
                  committer="me@foo.com",
                  comments="good stuff",
                  revision="HEAD",
                  project="none"
                  )

    @defer.inlineCallbacks
    def test_trigger(self):
        yield self.setupConfig(masterConfig())

        build = yield self.doForceBuild(wantSteps=True, useChange=self.change, wantLogs=True)

        self.assertEqual(
            build['steps'][2]['state_string'], 'triggered trigsched, 1 success')
        builds = yield self.master.data.get(("builds",))
        self.assertEqual(len(builds), 2)
        dump = StringIO()
        for b in builds:
            yield self.printBuild(b, dump)
        # depending on the environment the number of lines is different between
        # test hosts
        loglines = builds[1]['steps'][1]['logs'][0]['num_lines']
        self.assertRegex(dump.getvalue(),
                         expectedOutputRegex.format(loglines=loglines))

    @defer.inlineCallbacks
    def test_trigger_failure(self):
        yield self.setupConfig(masterConfig(addFailure=True))

        build = yield self.doForceBuild(wantSteps=True, useChange=self.change, wantLogs=True)

        self.assertEqual(
            build['steps'][2]['state_string'], 'triggered trigsched, 2 successes, 1 failure')
        builds = yield self.master.data.get(("builds",))
        self.assertEqual(len(builds), 4)


# master configuration
def masterConfig(addFailure=False):
    c = {}
    from buildbot.config import BuilderConfig
    from buildbot.process.factory import BuildFactory
    from buildbot.plugins import steps, schedulers

    c['schedulers'] = [
        schedulers.Triggerable(
            name="trigsched",
            builderNames=["build"]),
        schedulers.AnyBranchScheduler(
            name="sched",
            builderNames=["testy"])]

    f = BuildFactory()
    f.addStep(steps.ShellCommand(command='echo hello'))
    f.addStep(steps.Trigger(schedulerNames=['trigsched'],
                            waitForFinish=True,
                            updateSourceStamp=True))
    f.addStep(steps.ShellCommand(command='echo world'))
    f2 = BuildFactory()
    f2.addStep(steps.ShellCommand(command='echo ola'))
    c['builders'] = [
        BuilderConfig(name="testy",
                      workernames=["local1"],
                      factory=f),
        BuilderConfig(name="build",
                      workernames=["local1"],
                      factory=f2)]
    if addFailure:
        f3 = BuildFactory()
        f3.addStep(steps.ShellCommand(command='false'))
        c['builders'].append(BuilderConfig(name="build2", workernames=["local1"], factory=f3))
        c['builders'].append(BuilderConfig(name="build3", workernames=["local1"], factory=f2))
        c['schedulers'][0] = schedulers.Triggerable(name="trigsched",
                                                    builderNames=["build", "build2", "build3"])
    return c



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

ALL_RESULTS = list(range(7))
SUCCESS, WARNINGS, FAILURE, SKIPPED, EXCEPTION, RETRY, CANCELLED = ALL_RESULTS
Results = ["success", "warnings", "failure", "skipped", "exception", "retry", "cancelled"]
MultipleResults = ["successes", "warnings", "failures", "skipped", "exceptions",
                   "retries", "cancelled"]


def statusToString(status, count=1):
    if status is None:
        return "not finished"
    if status < 0 or status >= len(Results):
        return "Invalid status"
    if count > 1:
        return MultipleResults[status]
    return Results[status]


def worst_status(a, b):
    # SKIPPED > SUCCESS > WARNINGS > FAILURE > EXCEPTION > RETRY > CANCELLED
    # CANCELLED needs to be considered the worst.
    for s in (CANCELLED, RETRY, EXCEPTION, FAILURE, WARNINGS, SUCCESS, SKIPPED):
        if s in (a, b):
            return s
    return None


def computeResultAndTermination(obj, result, previousResult):
    possible_overall_result = result
    terminate = False
    if result == FAILURE:
        if not obj.flunkOnFailure:
            possible_overall_result = SUCCESS
        if obj.warnOnFailure:
            possible_overall_result = WARNINGS
        if obj.flunkOnFailure:
            possible_overall_result = FAILURE
        if obj.haltOnFailure:
            terminate = True
    elif result == WARNINGS:
        if not obj.warnOnWarnings:
            possible_overall_result = SUCCESS
        else:
            possible_overall_result = WARNINGS
        if obj.flunkOnWarnings:
            possible_overall_result = FAILURE
    elif result in (EXCEPTION, RETRY, CANCELLED):
        terminate = True

    result = worst_status(previousResult, possible_overall_result)
    return result, terminate


class ResultComputingConfigMixin:

    haltOnFailure = False
    flunkOnWarnings = False
    flunkOnFailure = True
    warnOnWarnings = False
    warnOnFailure = False

    resultConfig = [
        "haltOnFailure",
        "flunkOnWarnings",
        "flunkOnFailure",
        "warnOnWarnings",
        "warnOnFailure",
    ]



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.internet import task
from twisted.trial import unittest

from buildbot.schedulers import timed
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import scheduler


class Timed(scheduler.SchedulerMixin, TestReactorMixin, unittest.TestCase):

    OBJECTID = 928754

    def setUp(self):
        self.setup_test_reactor()
        self.setUpScheduler()

    def tearDown(self):
        self.tearDownScheduler()

    class Subclass(timed.Timed):

        def getNextBuildTime(self, lastActuation):
            self.got_lastActuation = lastActuation
            return defer.succeed((lastActuation or 1000) + 60)

        def startBuild(self):
            self.started_build = True
            return defer.succeed(None)

    def makeScheduler(self, firstBuildDuration=0, **kwargs):
        sched = self.attachScheduler(self.Subclass(**kwargs), self.OBJECTID)
        self.clock = sched._reactor = task.Clock()
        return sched
    # tests

    # note that most of the heavy-lifting for testing this class is handled by
    # the subclasses' tests, as that's the more natural place for it



#!/usr/bin/env python3

# this script takes all the PR created by dependabot and gather them into one

import argparse
import os

import requests
import yaml


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--remote', type=str, default=None,
                        help='The name of the remote to use for pull request. ' +
                             'Uses hub default if not specified')
    args = parser.parse_args()

    with open(os.path.expanduser("~/.config/hub")) as f:
        conf = yaml.safe_load(f)
        token = conf['github.com'][0]['oauth_token']

    os.system("git fetch https://github.com/buildbot/buildbot master")
    os.system("git checkout FETCH_HEAD -B gather_dependabot")
    s = requests.Session()
    s.headers.update({'Authorization': 'token ' + token})
    r = s.get("https://api.github.com/repos/buildbot/buildbot/pulls")
    r.raise_for_status()
    prs = r.json()
    with open("/tmp/hub_pr_message", 'w') as f:
        f.write("gather dependabot PRs\n\n")
        for pr in prs:
            if 'dependabot' in pr['user']['login']:
                print(pr['number'], pr['title'])
                f.write(f"#{pr['number']}: {pr['title']}\n")
                os.system(
                    "git fetch https://github.com/buildbot/buildbot "
                    f"refs/pull/{pr['number']}/head")
                os.system("git cherry-pick FETCH_HEAD")

    if args.remote is not None:
        os.system(f'git push {args.remote} gather_dependabot')
    os.system("hub pull-request -b buildbot:master -p -F /tmp/hub_pr_message -l dependencies")


if __name__ == '__main__':
    main()
